/* $EPIC:$ */
/*
 * symtab.c -- unified symbol table for epic
 *
 * Copyright  1999 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. All redistributions, whether of source code or in binary form must
 *    retain the the above copyright notice, the above paragraph (the one
 *    permitting redistribution), this list of conditions, and the following
 *    disclaimer.
 * 2. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

/*
 * A symbol, simply enough, is the union of an arbitrary name combined with
 * some data that is to be referenced when the symbol is used.  In our case,
 * we store all the data for all symbols in one place.  Each symbol may belong
 * to one or more "classes", which overlap.  
 *
 * The types of symbols:
 *
 *		ALIASes
 *		ASSIGNs
 *		Built in aliases
 *		Built in commands
 *		Built in functions
 *		SET variables
 *
 * The classes are:
 *
 *		VARIABLES (ASSIGNs, SETs,  Built in aliases, SETs, ENV vars)
 *		COMMANDS (ALIASes, Built in commands)
 *		FUNCTIONS (ALIASes, Built in functions)
 *
 * A symbol may be retreived specifically by type, or by class.  
 * A symbol may be looked up with and without completion. 
 * The data stored in a symbol is intentionally opaque.  When a symbol is
 * registered, the caller should not have to expose implementation details 
 * in order to store the data.  It is reasonable for a registered symbol's
 * data to be treated as a black box, which must then be passed to an
 * appropriate function to be processed (the function then would cast it.)
 */
struct symbol_type {
	char *		name;
	u_32int_t	hash;
	int		priority;
	void *		user_alias;
	void *		user_assign;
	void *		built_in_alias
	void *		built_in_command;
	void *		built_in_function;
	void *		built_in_set;
	void *		other;
};
typedef struct symbol_type Symbol;

/*
 * Of course, we need somewhere to store all of our symbols, so we stick them
 * in an alist.
 */
#define SYMBOL_CACHE_SIZE 4
typedef struct SymbolStru
{
	Symbol **	list;
	int		max;
	int		max_alloc;
	alist_func	func;
	hash_type	hash;
	Symbol **	cache;
	int		cache_size;
	int		revoke_index;
} SymbolTable;

static SymbolTable symtab = 	{ NULL, 0, 0, my_strnicmp,
					HASH_INSENSITIVE, NULL, 0, 0 };

/*
 * A "symbol operation" describes a symbol name and an action to take on
 * that symbol, whether it be addition or subtraction or what have you.
 */
enum SymbolOp {
	SYMBOL_ADD,
	SYMBOL_DELETE,
	SYMBOL_GET
};
enum SymbolType {
	USER_ALIAS,
	USER_ASSIGN,
	BUILT_IN_ALIAS,
	BUILT_IN_COMMAND,
	BUILT_IN_FUNCTION,
	BUILT_IN_SET,
	OTHER,
	COMMAND_GLOB,
	VARIABLE_GLOB,
	FUNCTION_GLOB
};

typedef struct SymbolOpStru 
{
	char *		name;
	int		priority;
	SymbolType	type;
	SymbolOp	operation;
	char *		input_data;
	char *		output_data;
} SymbolOpSet

int	symbolctl (int numops, SymbolOpSet *operations)
{
	int		i;
	SymbolOpSet *	op;
	void **		ptr;

	for (i = 0; i < numops; i++)
	{
	    op = &operations[i];
	    op->output = NULL;

	    switch (op->operation)
	    {
		case SYMBOL_ADD:
		{
			symbol = get_symbol_create(op->name);
			switch (op->type) 
			{
			    case USER_ALIAS:
				op->output = symbol->user_alias;
				symbol->user_alias = op->input;
				break;
			    case USER_ASSIGN:
				op->output = symbol->user_assign;
				symbol->user_assign = op->input;
				break;
			    case BUILT_IN_ALIAS:
				op->output = symbol->built_in_alias;
				symbol->built_in_alias = op->input;
				break;
			    case BUILT_IN_COMMAND:
				op->output = symbol->built_in_command;
				symbol->built_in_command = op->input;
				break;
			    case BUILT_IN_FUNCTION:
				op->output = symbol->built_in_function;
				symbol->built_in_function = op->input;
				break;
			    case BUILT_IN_SET:
				op->output = symbol->built_in_set;
				symbol->built_in_set = op->input;
				break;
			}
			break;
		}

		case SYMBOL_DELETE:
		{
			if (!(symbol = get_symbol(op->name)))
				break;

			switch (op->type) 
			{
			    case USER_ALIAS:
				op->output = symbol->user_alias;
				symbol->user_alias = NULL;
				break;
			    case USER_ASSIGN:
				op->output = symbol->user_assign;
				symbol->user_assign = NULL;
				break;
			    case BUILT_IN_ALIAS:
				op->output = symbol->built_in_alias;
				symbol->built_in_alias = NULL;
				break;
			    case BUILT_IN_COMMAND:
				op->output = symbol->built_in_command;
				symbol->built_in_command = NULL;
				break;
			    case BUILT_IN_FUNCTION:
				op->output = symbol->built_in_function;
				symbol->built_in_function = NULL;
				break;
			    case BUILT_IN_SET:
				op->output = symbol->built_in_set;
				symbol->built_in_set = NULL;
				break;
			}
			check_symbol(symbol);
			break;
		}

		case SYMBOL_GET:
		{
			/* XXX This doesn't do completion which it must */
			if (!(symbol = get_symbol(op->name)))
				break;

			switch (op->type) 
			{
			    case USER_ALIAS:
				op->output = symbol->user_alias;
				break;
			    case USER_ASSIGN:
				op->output = symbol->user_assign;
				break;
			    case BUILT_IN_ALIAS:
				op->output = symbol->built_in_alias;
				break;
			    case BUILT_IN_COMMAND:
				op->output = symbol->built_in_command;
				break;
			    case BUILT_IN_FUNCTION:
				op->output = symbol->built_in_function;
				break;
			    case BUILT_IN_SET:
				op->output = symbol->built_in_set;
				break;

			    /* These are "meta" types */
			    case COMMAND_GLOB:
				if (symbol->user_alias)
					op->output = symbol->user_alias;
				else if (symbol->built_in_command)
					op->output = symbol->built_in_command;
				break;
			    case VARIABLE_GLOB:
				if (symbol->user_assign)
					op->output = symbol->user_assign;
				else if (symbol->built_in_alias)
					op->output = symbol->built_in_alias;
				break;
			    case FUNCTION_GLOB:
				if (symbol->built_in_function)
					op->output = symbol->built_in_function;
				else if (symbol->user_alias)
					op->output = symbol->user_alias;
				break;
			}
			break;
		}
	     }
	}
}

Symbol *make_new_Symbol (char *name)
{
	Symbol *tmp = (Symbol *)new_malloc(sizeof(Symbol));
	tmp->name = m_strdup(name);
	tmp->priority = current_priority;
	tmp->user_alias = NULL;
	tmp->user_assign = NULL;
	tmp->built_in_alias = NULL;
	tmp->built_in_command = NULL;
	tmp->built_in_function = NULL;
	tmp->built_in_set = NULL;
	tmp->other = NULL;
	return tmp;
}

Symbol *get_symbol (const char *name)
{
	int		cnt, loc;
	Symbol *	tmp;

	/* Check the cache */
	tmp = (Symbol *)find_array_item((array *)&symtab, name, &cnt, &loc);
	if (!tmp || cnt >= 0)
		return NULL;
	return tmp;
}

Symbol *get_symbol_create (const char *name)
{
	Symbol *	tmp;

	if (!(tmp = get_symbol(name)))
		return make_new_symbol(name);
}

/* Obviously i havent gotten much further than this. */
