/* $EPIC:$ */
/*
 * names.c: This here is used to maintain a list of all the people currently
 * on your channel.  Seems to work 
 *
 * Copyright (c) 1990 Michael Sandroff.
 * Copyright (c) 1991, 1992 Troy Rollo.
 * Copyright (c) 1992-1996 Matthew Green.
 * Copyright  1993-1999 Jeremy Nelson and others ("EPIC Software Labs").
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notices, the above paragraph (the one permitting redistribution),
 *    this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the author(s) may not be used to endorse or promote 
 *    products derived from this software without specific prior written
 *    permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 */

#define NEED_SERVER_LIST
#include "irc.h"
#include "alist.h"
#include "names.h"
#include "ircaux.h"
#include "server.h"
#include "output.h"
#include "screen.h"
#include "window.h"
#include "debug.h"
#include "vars.h"
#include "timer.h"

void 	add_user_who (char *from, char **ArgList);
void 	add_user_end (char *from, char **ArgList);
char *	make_friendly (Char *channel);

#define __inline__

/*
 * NOTE TO FUTURE MAINTAINERS!
 *
 * This file was written in an object oriented fashion as a proof of concept
 * work.  Immediately following each structure is a list of public functions
 * that operate on that structure.  You may only call these functions to
 * manipulate data in that structure!  If you find that you need to do
 * something that is not covered by the functions, then write a new function
 * to do it.  PLEASE DO NOT DIRECTLY ACCESS THE DATA IN THE STRUCTURES!
 * This is for the protection of the data as well as for your sanity.
 */


/****************************** NICKS ************************************/
/*
 * A nick item is pretty simple -- it stores the nickname, the userhost,
 * the channel op status, the channel voice status, and the channel helper
 * status.
 */
typedef struct nick_stru
{
	char *	nick;		/* nickname of person on channel */
	u_32int_t hash;		/* Hash of the nickname */
	char *	userhost;	/* Their userhost, if we know it */
	int	suspicious;	/* 1 if it is, 0 if it is not */
	int	chanop;		/* True if they are a channel operator */
	int	voice;		/* 1 if they are, 0 if theyre not, -1 if uk */
	int	half_assed;	/* 1 if they are, 0 if theyre not, -1 if uk */
}	Nick;

static	Nick *	Nick__new (Char *nick, Char *uh, int susp, int chop, int voice, int half);
static	void	Nick__delete (Nick *this);
static	void	Nick__nickname_set (Nick *this, Char *nick);
static	Char *	Nick__nickname_get (Nick *this);
static	void	Nick__userhost_set (Nick *this, Char *uh);
static	Char *	Nick__userhost_get (Nick *this);
static	void	Nick__chanop_set (Nick *this, int value);
static	int	Nick__chanop_get (Nick *this);
static	void	Nick__voice_set (Nick *this, int value);
static	int	Nick__voice_get (Nick *this);
static	void	Nick__helper_set (Nick *this, int value);
static	int	Nick__helper_get (Nick *this);

/* * * IMPLEMENTATION * * */
__inline__
static	Nick *	Nick__new (Char *nick, Char *uh, int susp, int chop, int voice, int half)
{
	Nick *	this;

	this = (Nick *)new_malloc(sizeof(Nick));
	this->nick = m_strdup(nick);
	this->userhost = m_strdup(uh);
	this->chanop = chop;
	this->voice = voice;
	this->half_assed = half;
	this->suspicious = susp;
	return this;
}

__inline__
static	void	Nick__delete (Nick *this)
{
	new_free(&this->nick);
	new_free(&this->userhost);
	new_free((char **)&this);
}

__inline__
static	void	Nick__nickname_set (Nick *this, Char *nick)
{
	malloc_strcpy(&this->nick, nick);
}

__inline__
static	Char *	Nick__nickname_get (Nick *this)
{
	return this->nick;
}

__inline__
static	void	Nick__userhost_set (Nick *this, Char *uh)
{
	malloc_strcpy(&this->userhost, uh);
}

__inline__
static	Char *	Nick__userhost_get (Nick *this) 
{
	return this->userhost;
}

__inline__
static	void	Nick__chanop_set (Nick *this, int value) 
{
	this->chanop = value;
}

__inline__
static	int	Nick__chanop_get (Nick *this)
{
	return this->chanop;
}

__inline__
static	void	Nick__voice_set (Nick *this, int value)
{
	this->voice = value;
}

__inline__
static	int	Nick__voice_get (Nick *this)
{
	return this->voice;
}

__inline__
static	void	Nick__helper_set (Nick *this, int value)
{
	this->half_assed = value;
}

__inline__
static	int	Nick__helper_get (Nick *this)
{
	return this->half_assed;
}

__inline__
static	int	Nick__suspicious_get (Nick *this)
{
	return this->suspicious;
}


/*************************** NICK LISTS *********************************/
/*
 * In each channel is a "nick list", which is just an alist of Nick 
 * structures.  Nothing very fancy.
 */
typedef	struct	nick_alist_stru
{
	Nick **		list;
	int		max;
	int		max_alloc;
	alist_func 	func;
	hash_type 	hash;
}	NickList;

static	NickList *	NickList__new (void);
static	void		NickList__delete (NickList *this);
static	void		NickList__addNick (NickList *this, Nick *item);
static	Nick *		NickList__removeNick (NickList *this, Nick *item);
static	Nick *		NickList__getNick (NickList *this, Char *nick);
static	Nick *		NickList__getN (NickList *this, int which);
static	int		NickList__get_size (NickList *this);

/* * * IMPLEMENTATION * * */
__inline__
static	NickList *	NickList__new (void)
{
	NickList *this;

	this = (NickList *)new_malloc(sizeof(NickList));
	this->list = NULL;
	this->max = 0;
	this->max_alloc = 0;
	this->func = strncmp;
	this->hash = HASH_INSENSITIVE;
	return this;
}

__inline__
static	void		NickList__delete (NickList *this)
{
	int	i;

	for (i = 0; i < this->max; i++)
		Nick__delete(this->list[i]);
	new_free((char **)&this->list);
	new_free((char **)&this);
}

__inline__
static	void		NickList__addNick (NickList *this, Nick *item)
{
	Nick *	old;
	if ((old = (Nick *)add_to_array((array *)this, (array_item *)item)))
		Nick__delete(old);
}

__inline__
static	Nick *		NickList__removeNick (NickList *this, Nick *item)
{
	/* XXX Inefficient. */
	if (!remove_from_array((array *)this, item->nick))
		return NULL;
	return item;
}

__inline__
static	Nick *		NickList__getNick (NickList *this, Char *nick)
{
	Nick *	item;
	int	count;
	int	loc;

	item = (Nick *)find_array_item((array *)this, nick, &count, &loc);
	if (count < 0)
		return item;
	else
		return NULL;
}

__inline__
static	Nick *		NickList__getN (NickList *this, int which)
{
	if (which >= 0 && which < this->max)
		return this->list[which];
	return NULL;
}

__inline__
static	int		NickList__get_size (NickList *this)
{
	return this->max;
}


/******************************** CHANNELS *********************************/
typedef enum	channel_priv_u {
	NOT_CURRENT,
	CURRENT
} 	CCurrent;

typedef	struct	channel_stru
{
	Chname		channel;	/* The channel's canonical name */
	u_32int_t	hash;		/* Hash of the channel name */

	Chname		friendly;	/* What we tell the user the name is */
	Servref		server;		/* last or present server connection */
	Winref		window;		/* the window the channel is "on" */
	CStatus		status;		/* Operation status of channel */
	CCurrent	current;	/* Current channel for window? */
	int		rejoin_delay;	/* How many seconds to wait. */
	NickList *	nicks;		/* alist of nicks on channel */

	u_long		mode;		/* Current mode settings for channel */
	u_long		i_mode;		/* channel mode for cached string */
	char *		s_mode;		/* cached string version of modes */
	char *		saved_mode;	/* Cached string for rejoin */
	int		limit;		/* max users for the channel */
	char *		key;		/* key for this channel */

	char		chop;		/* true if I am a channel operator */
	char		voice;		/* true if I am voiced */
	char		half_assed;	/* true if I am a half-assed oper */
}	Channel;

static	Channel *	Channel__new (Chname name, Servref server, Winref w);
static	void		Channel__delete (Channel *this);
static 	void		Channel__join (Channel *this);
static 	void		Channel__part (Channel *this, Char *reason);
static 	void		Channel__status_change (Channel *this, CStatus newc);
static 	CStatus		Channel__status_get (Channel *this);
static 	void		Channel__current_change (Channel *this, CCurrent newc);
static 	CCurrent 	Channel__current_get (Channel *this);
static	NickList *	Channel__nicklist (Channel *this);
static	u_long		Channel__mode_logical (Channel *this);
static	void		Channel__mode_change (Channel *this, Char *modes);
static	char *		Channel__mode_string (Channel *this);
static	int		Channel__limit (Channel *this);
static	Char *		Channel__key (Channel *this);
static	void		Channel__key_set (Channel *this, Char *newkey);
static	int		Channel__im_chop (Channel *this);
static	int		Channel__im_voice (Channel *this);
static	int		Channel__im_helper (Channel *this);
static	int		Channel__private (Channel *this);
static	int		Channel__nomsgs (Channel *this);
static	Winref		Channel__window_get (Channel *this);
static	void		Channel__window_set (Channel *this, Winref w);
static	Servref		Channel__server_get (Channel *this);
static	void		Channel__server_set (Channel *this, Servref s);

/* * * IMPLEMENTATION * * */
__inline__
static	Channel *	Channel__new (Chname name, Servref server, Winref w)
{
	Channel *this = (Channel *)new_malloc(sizeof(Channel));
	if (x_debug & DEBUG_CHANNELS)
		yell("Creating a new channel, [%s] [%d] [%d]", name, server, w);

	this->channel = m_strdup(name);
	this->friendly = make_friendly(name);
	this->server = server;
	this->window = w;
	this->status = UNKNOWN;
	this->current = NOT_CURRENT;
	this->nicks = NickList__new();
	this->mode = 0;
	this->rejoin_delay = -1;
	this->i_mode = 0;
	this->s_mode = NULL;
	this->saved_mode = NULL;
	this->limit = 0;
	this->key = NULL;
	this->chop = 0;
	this->voice = 0;
	this->half_assed = 0;
	return this;
}

__inline__
static	void		Channel__delete (Channel *this)
{
	if (x_debug & DEBUG_CHANNELS)
		yell("Final cleanup for channel [%s] [%d] [%d]", 
			this->channel, this->server, this->window);

	new_free(&this->channel);
	new_free(&this->friendly);
	NickList__delete(this->nicks);
	this->nicks = NULL;		/* Just in case */
	new_free(&this->s_mode);
	new_free(&this->saved_mode);
	new_free(&this->key);
}

__inline__
static 	void		Channel__join (Channel *this)
{
	if (!this)
		panic("Channel__join: this == NULL");

	if (x_debug & DEBUG_CHANNELS)
		yell("Joining channel [%s] [%d] [%d]", 
			this->channel, this->server, this->window);

	if (this->status >= JOIN_SENT && this->status < PART_SENT)
	{
		say("Cannot join %s, already on channel", this->channel);
		return;
	}

	if (this->key)
		send_to_aserver(this->server, "JOIN %s %s", 
				this->channel, this->key);
	else
		send_to_aserver(this->server, "JOIN %s", this->channel);

	Channel__status_change(this, JOIN_SENT);
}

__inline__
static 	void		Channel__part (Channel *this, Char *reason)
{
	if (!this)
		panic("Channel__part: this == NULL");

	if (x_debug & DEBUG_CHANNELS)
		yell("Parting channel [%s] [%d] [%d]", 
			this->channel, this->server, this->window);

	if (this->status >= JOIN_SENT && this->status < PART_SENT)
	{
	    if (reason)
		send_to_aserver(this->server, "PART %s :%s", 
					this->channel, reason);
	    else
		send_to_aserver(this->server, "PART %s", this->channel);
	}
	else
	{
		say("Cannot leave %s, not on channel", this->channel);
		return;
	}

	Channel__status_change(this, PART_SENT);
}

static 	void		Channel__status_change (Channel *this, CStatus newc)
{
	CStatus	old;

	old = this->status;
	this->status = newc;

	if (x_debug & DEBUG_CHANNELS)
	    yell("Switching channel status [%s] [%d] [%d], from [%d] to [%d]",
		this->channel, this->server, this->window,
		old, this->status);

	switch (this->status)
	{
	    /*
	     * This means something bad happened, like the join was refused.
	     */
	    case (UNKNOWN):
	    {
		panic("Unimplemnted");
		/* Tell the user we give up. */
		break;
	    }

	    /*
	     * This mode is when Channel__join is called and a new JOIN
	     * request is sent out for this channel.
	     */
	    case (JOIN_SENT):
	    {
		/* Throw away the old channel list */
		NickList__delete(this->nicks);
		this->nicks = NickList__new();

		/* Save the old channel mode so we can restore it later. */
		yell("Channel mode restoration not supported yet.");
		break;
	    }

	    /*
	     * This mode is when the JOIN protocol command has been recieved
	     * and it is time to start syncing the channel up.
	     */
	    case (SYNCING):
	    {
		char *value = LOCAL_COPY(this->channel);

		if (old != JOIN_SENT)
			yell("Unexpected JOIN recieved.  How rude.");

		/* Ask for the channel's mode */
		send_to_aserver(this->server, "MODE %s", this->channel);

		/* Ask for info about the channel's users */
		whobase(value, add_user_who, add_user_end);

		break;
	    }

	    /*
	     * This mode is when all of the stateful data about the channel
	     * has been fully recievced and processed and we are ready to
	     * go into normal cruising mode.
	     */
	    case (ACTIVE):
	    {
		/* Hook /on channel_sync or whatever. */
		break;
	    }

	    /*
	     * This mode is when Channel__part is called and a PART request
	     * has been sent out for this channel.  Whether or not the channel
	     * will be destroyed depends on the prior state.
	     */
	    case (PART_SENT):
	    {
		/* Blow away current channel status */
		is_current_channel(this->channel, 1);
		/* Nothing to do, really */
		break;
	    }

	    /*
	     * This mode when p_kick notices we've been kicked from a channel.
	     * This is responsible for doing whatever is neccesary to handle
	     * that situation.
	     */
	    case (KICKED):
	    {
		if (this->rejoin_delay == -1 && !get_int_var(AUTO_REJOIN_VAR))
			goto do_deceased;

		add_timer(0, empty_string,
			get_int_var(AUTO_REJOIN_DELAY_VAR) * 1000,
			1, auto_rejoin_callback,
			m_sprintf("%s %d", this->channel, this->server),
			NULL, -1);

		/* Plus whatever else is neccesary. */
		break;
	    }

	    /*
	     * This mode is when a server has been disconnected and some or
	     * all of its channels have not been rejoined and so they wait
	     * to be rejoined on the original server.
	     */
	    case (DISCONNECTED):
	    {
		panic("Unimplemnted");
		/* Delete the users on the channel */
		break;
	    }

	    /*
	     * While a server reconnect is pending, all of the channels on
	     * that server are locked.  If a new server is connected to,
	     * then these channels are transfered to the new server and
	     * unlocked.  If a new server is not connected to, then this
	     * channel will go to DISCONNECTED mode.
	     */
	    case (LOCKED):
	    {
		panic("Unimplemnted");
		/* Delete the users on the channel */
		break;
	    }

	    /*
	     * Typically, this is when a PART has been recieved for this
	     * channel.  Depending on the prior mode, this may cause the
	     * channel to be destroyed or to be immediately rejoined.
	     */
	    case (DECEASED):
	    {
do_deceased:
		/* XXX Cough.  Ick. this is bogus. XXX */
		if (old != PART_SENT)
			yell("WARNING!  Channel [%s:%d] being removed forever",
				this->channel, this->server);
		channel_remove(this->channel, this->server);
		this = NULL;
		break;
	    }
	}
}

__inline__
static 	CStatus		Channel__status_get (Channel *this)
{
	return this->status;
}

__inline__
static 	void		Channel__current_change (Channel *this, CCurrent newc)
{
	this->current = newc;
}

__inline__
static 	CCurrent 	Channel__current_get (Channel *this)
{
	return this->current;
}

__inline__
static	NickList *	Channel__nicklist (Channel *this)
{
	return this->nicks;
}

__inline__
static	u_long		Channel__mode_logical (Channel *this)
{
	return this->i_mode;
}

__inline__
static	void		Channel__mode_change (Channel *this, Char *modes)
{
	char *	limit = 0;
	Char *	person;
	int	add = 0;
	char *	rest;
	Nick *	nick;
	int	value = 0;
	char *	mode_str;

	mode_str = LOCAL_COPY(modes);

	/* Punt if its not all there */
	if (!(mode_str = next_arg(mode_str, &rest)))
		return;

	for (; *mode_str; mode_str++)
	{
	    switch (*mode_str)
	    {
		case '+':
			add = 1;
			continue;
		case '-':
			add = 0;
			continue;

		case 'i':
			value = MODE_INVITE;
			break;
		case 'm':
			value = MODE_MODERATED;
			break;
		case 'n':
			value = MODE_MSGS;
			break;
		case 'p':
			value = MODE_PRIVATE;
			break;
		case 'r':
			value = MODE_REGISTERED;
			break;
		case 'R':
			value = MODE_RESTRICTED;
			break;
		case 's':
			value = MODE_SECRET;
			break;
		case 't':
			value = MODE_TOPIC;
			break;
		case 'z':		/* Erf/TS4 "zapped" */
			value = MODE_Z;
			break;

		case 'k':
		{
			char *key = next_arg(rest, &rest);

			value = MODE_KEY;

			if (add)
				malloc_strcpy(&this->key, key);
			else
				new_free(&this->key);

			this->i_mode = -1;	/* Revoke old cache */
			break;	
		}

		case 'l':
		{
			value = MODE_LIMIT;
			if (add)
				limit = next_arg(rest, &rest);
			else
				limit = zero;

			this->limit = my_atol(limit);
			this->i_mode = -1;	/* Revoke old cache */
			continue;
		}

		case 'o':
		{
			/* 
			 * Borked av2.9 sends a +o to the channel
			 * when you create it, but doesnt bother to
			 * send your nickname, too. blah.
			 */
			if (!(person = next_arg(rest, &rest)))
			    person = get_server_nickname(from_server);

			nick = NickList__getNick(this->nicks, person);
			if (!nick)
			{
				say("MODE %s +o %s doesn't make any sense", 
					this->channel, person);
			}
			else
			{
				nick->chanop = add;
				if (is_me(from_server, person))
					this->chop = add;
			}
			continue;
		}

		case 'v':
		{
			if (!(person = next_arg(rest, &rest)))
			   person = get_server_nickname(from_server);

			nick = NickList__getNick(this->nicks, person);
			if (!nick)
			{
				say("MODE %s +v %s doesn't make any sense", 
					this->channel, person);
			}
			else
			{
				nick->voice = add;
				if (is_me(from_server, person))
					this->voice = add;
			}
			continue;
		}

		case 'h': /* erfnet's borked 'half-assed oper' mode */
		{
			if (!(person = next_arg(rest, &rest)))
			   person = get_server_nickname(from_server);

			nick = NickList__getNick(this->nicks, person);
			if (!nick)
			{
				say("MODE %s +h %s doesn't make any sense", 
					this->channel, person);
			}
			else
			{
				nick->half_assed = add;
				if (is_me(from_server, person))
					this->half_assed = add;
			}
			continue;
		}

		case 'b':
		case 'e':	/* borked erfnet ban exceptions */
		{
			next_arg(rest, &rest);
			continue;
		}
	    }

	    if (add)
		this->mode |= value;
	    else
		this->mode &= ~value;
	}

	if (this->limit == 0)
		this->mode &= ~MODE_LIMIT;
	else
		this->mode |= MODE_LIMIT;
}

/*
 * Channel__mode_string: Get the current mode string for the specified 
 * server.  We only do a refresh if the mode has changed since we 
 * generated it.
 */
static char *	Channel__mode_string (Channel *this)
{
	int	mode_pos = 0,
		str_pos = 0;
	char	local_buffer[BIG_BUFFER_SIZE];

	/* Used the cache value if its still good */
	if ((this->mode == this->i_mode) && this->s_mode)
		return this->s_mode;

	this->i_mode = this->mode;

	local_buffer[0] = 0;
	while (this->mode >= (1 << mode_pos))
	{
		if (this->mode & (1 << mode_pos))
			local_buffer[str_pos++] = cmodes[mode_pos];
		mode_pos++;
	}
	local_buffer[str_pos] = 0;

	if (this->key)
	{
		strlcat(local_buffer, " ", BIG_BUFFER_SIZE);
		strlcat(local_buffer, this->key, BIG_BUFFER_SIZE);
	}
	if (this->limit)
	{
		strlcat(local_buffer, " ", BIG_BUFFER_SIZE);
		strlcat(local_buffer, ltoa(this->limit), BIG_BUFFER_SIZE);
	}

	malloc_strcpy(&this->s_mode, local_buffer);
	return (this->s_mode);
}

__inline__
static	int		Channel__limit (Channel *this)
{
	return this->limit;
}

__inline__
static	Char *		Channel__key (Channel *this)
{
	return this->key;
}

__inline__
static	void		Channel__key_set (Channel *this, Char *newkey)
{
	if (newkey)
		malloc_strcpy(&this->key, newkey);
	else
		new_free(&this->key);
}

__inline__
static	int		Channel__im_chop (Channel *this)
{
	return this->chop;
}

__inline__
static	int		Channel__im_voice (Channel *this)
{
	return this->voice;
}

__inline__
static	int		Channel__im_helper (Channel *this)
{
	return this->half_assed;
}

__inline__
static	int		Channel__private (Channel *this)
{
	return (this->i_mode & MODE_PRIVATE) ? 1 : 0;
}

__inline__
static	int		Channel__nomsgs (Channel *this)
{
	return (this->i_mode & MODE_MSGS) ? 1 : 0;
}

__inline__
static void		Channel__show (Channel *this)
{
	NickList *	nicks = this->nicks;
	Nick *		nick = NULL;
	char		local_buf[BIG_BUFFER_SIZE * 10 + 1];
	char *		ptr;
	int		nick_len;
	int		len;
	int		i;

	ptr = local_buf;
	*ptr = 0;
	nick_len = BIG_BUFFER_SIZE * 10;

	for (i = 0; i < NickList__get_size(nicks); i++)
	{
		nick = NickList__getN(nicks, i);
		strlcpy(ptr, Nick__nickname_get(nick), nick_len);
		if (Nick__userhost_get(nick))
		{
			strlcat(ptr, "!", nick_len);
			strlcat(ptr, Nick__userhost_get(nick), nick_len);
		}
		strlcat(ptr, space, nick_len);

		len = strlen(ptr);
		nick_len -= len;
		ptr += len;

		if (nick_len <= 0)
			break;		/* No more space */
	}

	say("\t%s +%s (%s) (Win: %d): %s", 
		this->channel, 
		Channel__mode_string(this),
		get_server_name(this->server), 
		this->window, local_buf);
}

__inline__
static	Winref		Channel__window_get (Channel *this)
{
	return this->window;
}

__inline__
static	void		Channel__window_set (Channel *this, Winref w)
{
	if (this->server == get_window_server(w))
		this->window = w;
	else
		panic("This channel is not on that server");
}

__inline__
static	Servref		Channel__server_get (Channel *this)
{
	return this->server;
}

__inline__
static	void		Channel__server_set (Channel *this, Servref s)
{
	this->server = s;
	this->window = -1;	/* Suspend the window for now */
}


/************************** CHANNEL LISTS ********************************/
/*
 * And for each server, it has its own private list of channels.  For what
 * I hope are obvious reasons, channel names are supposed to be unique in
 * each server connection.  If a server recieves a new channel that already
 * exists, that new channel should be dropped.  Nothing fancy here.
 */
typedef struct ChannelListStru
{
	Channel **	list;
	int		max;
	int		max_alloc;
	alist_func	func;
	hash_type	hash;
}	ChannelList;

static	ChannelList *	ChannelList__new (void);
static	void		ChannelList__delete (ChannelList *this);
static	void		ChannelList__addChannel (ChannelList *this, 
						 Channel *item);
static	Channel *	ChannelList__removeChannel (ChannelList *this, 
						    Channel *item);
static	Channel *	ChannelList__getChannel (ChannelList *this, 
						 Char *name);
static	void		ChannelList__traverse (ChannelList *this, void *magic, 
					       int (*func) (Channel *, void *));
static	Channel *	ChannelList__getN (ChannelList *this, int n);

/* * * Implementation * * */
__inline__
static	ChannelList *	ChannelList__new (void)
{
	ChannelList *this;

	this = (ChannelList *)new_malloc(sizeof(ChannelList));
	this->list = NULL;
	this->max = 0;
	this->max_alloc = 0;
	this->func = strncmp;
	this->hash = HASH_INSENSITIVE;
	return this;
}

__inline__
static	void	ChannelList__delete (ChannelList *this)
{
	int	i;

	for (i = 0; i < this->max; i++)
		Channel__delete(this->list[i]);
	new_free((char **)&this->list);
	new_free((char **)&this);
}

__inline__
static	void	ChannelList__addChannel (ChannelList *this, Channel *item)
{
	Channel *old;
	if ((old = (Channel *)add_to_array((array *)this, 
					   (array_item *)item)))
		Channel__delete(old);
}

__inline__
static	Channel *ChannelList__removeChannel (ChannelList *this, Channel *item)
{
	/* XXX Inefficient */
	if (!remove_from_array((array *)this, item->channel))
		return NULL;
	return item;
}

__inline__
static	Channel *ChannelList__getChannel (ChannelList *this, Char *name)
{
	Channel *item;
	int	count;
	int	loc;

	item = (Channel *)find_array_item((array *)this, 
						name, &count, &loc);
	if (count < 0)
		return item;
	else
		return NULL;
}

__inline__
static void	ChannelList__traverse (ChannelList *this, void *magic, int (*func) (Channel *, void *))
{
	int	i;

	for (i = 0; i < this->max; i++)
		if (func(this->list[i], magic))
			break;
}

__inline__
static	Channel *ChannelList__getN (ChannelList *this, int n)
{
	if (n >= 0 && n < this->max)
		return this->list[n];
	return NULL;
}

ChannelList *init_channels (void)
{
	return ChannelList__new();
}



/************************* INTERNAL FUNCTIONS ***************************/
__inline__
static Channel *get_server_channel (Chname channel_name, Servref server)
{
	return ChannelList__getChannel(get_server_channels(server),
					channel_name);
}

__inline__
static Nick *	get_server_channel_nick (Chname channel, Servref server, Char *nickn)
{
	Channel	*chan;
	Nick *	nick;

	if ((chan = get_server_channel(channel, server)))
		if ((nick = NickList__getNick(chan->nicks, nickn)))
			return nick;

	return NULL;
}

__inline__
static	Channel *traverse_all_channels (int *n, Servref server)
{
	ChannelList *cl = get_server_channels(server);

	(*n)++;
	return ChannelList__getN(cl, *n);
}



/************************ EXPORTED FUNCTIONS **************************/
void	channel_create	(Chname channel, Servref server, Winref w, Char *key)
{
	ChannelList *cl = get_server_channels(server);
	Channel *ch = Channel__new(channel, server, w);

	ChannelList__addChannel(cl, ch);
	Channel__key_set(ch, key);
	Channel__join(ch);
}

void	channel_remove	(Chname channel, Servref server)
{
	ChannelList *cl;
	Channel *ch;

	if (server == -1)
	{
		int	count;
		for (count = 0; count < number_of_servers; count++)
			channel_remove(NULL, count);
		return;
	}

	if (!channel)
	{
		channel_server_destroy(server);
		return;
	}

	cl = get_server_channels(server);
	ch = ChannelList__getChannel(cl, channel);
	ChannelList__removeChannel(cl, ch);
	Channel__delete(ch);
}

int 	channel_im_on (Chname channel, Servref server)
{
	return (get_server_channel(channel, server)) ? 1 : 0;
}

void 	channel_mode_update (Chname channel, Servref server, Char *mode)
{
	Channel *ch = get_server_channel(channel, server);

	if (ch)
		Channel__mode_change(ch, mode);
}

void	channel_status_change (Chname channel, Servref server, CStatus news)
{
	Channel *ch = get_server_channel(channel, server);
	if (ch)
		Channel__status_change(ch, news);
}

void	channel_suspend (Servref server)
{
	panic("Unimplemented");
}

Char *	channel_friendly (Chname channel, Servref server)
{
	Channel *ch;

	if (server == -1)
		server = from_server;
	if ((ch = get_server_channel(channel, server)))
		return ch->friendly;
	return empty_string;
}

Char *	channel_key (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	Char	*retval;

	if (ch && (retval = Channel__key(ch)))
		return retval;
	return empty_string;
}

Char	*channel_mode (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	Char	*retval;

	if (ch && (retval = Channel__mode_string(ch)))
		return retval;
	return empty_string;
}

/*
 * is_channel_mode: returns the logical AND of the given mode with the
 * channels mode.  Useful for testing a channels mode 
 */
int 	channel_private (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	u_long	mode;

	if (ch && (mode = Channel__mode_logical(ch)))
		return mode & (MODE_PRIVATE | MODE_SECRET);
	return 0;
}

int	channel_nomsgs (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	u_long	mode;

	if (ch && (mode = Channel__mode_logical(ch)))
		return mode & (MODE_MSGS);
	return 0;
}

/* list_channels: displays your current channel and your channel list */
void 	channel_list (void)
{
	Channel *tmp = NULL;
	int	count = -1;

	if (!channel_list)
	{
		say("You are not on any channels");
		return;
	}

	if (get_window_channel(0))
		say("Current channel %s", get_window_channel(0));
	else
		say("No current channel for this window");

	say("You are on the following channels:");
	while ((tmp = traverse_all_channels(&count, from_server)))
		Channel__show(tmp);

	if (connected_to_server != 1)
	{
		int i;

		say("Other servers:");
		for (i = 0; i < number_of_servers; i++)
		{
		    count = -1;
		    if (i != from_server)
			while ((tmp = traverse_all_channels(&count, i)))
				Channel__show(tmp);
		}
	}
}

int	channel_im_oper (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	if (ch)
		return Channel__im_chop(ch);
	return 0;
}

int	channel_im_voice (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	if (ch)
		return Channel__im_voice(ch);
	return 0;
}

void	channel_switch (char unused, char *ignored)
{
	panic("Unimplemented");
}

void	channel_server_change (Servref oldserver, Servref newserver)
{
	ChannelList *old;
	ChannelList *new;
	Channel *ch;

	if (oldserver == -1 || newserver == -1)
		return;			/* Whatever */

	if (!(old = get_server_channels(oldserver)))
		return;

	if (!(new = get_server_channels(newserver)))
		return;

	while ((ch = ChannelList__getN(old, 0)))
	{
		ChannelList__removeChannel(old, ch);
		Channel__server_set(ch, newserver);
		ChannelList__addChannel(new, ch);
	}
}

void	channel_server_remove (Servref server)
{
	panic("Unimplemented");
}

void	channel_server_reconnect (Servref server)
{
	ChannelList *cl = get_server_channels(server);
	Channel *ch;

	while ((ch = ChannelList__getN(cl, 0)))
		Channel__join(ch);
}

void	channel_server_destroy (Servref server)
{
	ChannelList *cl = get_server_channels(server);
	Channel *ch;

	while ((ch = ChannelList__getN(cl, 0)))
	{
		ChannelList__removeChannel(cl, ch);
		Channel__delete(ch);
	}
}

Char *	channel_what (Nickname nick)
{
	Channel *tmp = NULL;
	int	count = -1;

	if (current_window->current_channel &&
	    channel_nick_onchannel(current_window->current_channel, 
				   current_window->server, nick))
		return current_window->current_channel;

	while ((tmp = traverse_all_channels(&count, from_server)))
	{
		if (NickList__getNick(tmp->nicks, nick))
			return tmp->channel;
	}

	return NULL;
}

int	channel_getlist (Servref server, Nickname nick, Channel_List list)
{
	Channel *	tmp;
	int		count = -1;
	int		idx = 0;

	while ((tmp = traverse_all_channels(&count, server)))
		if (NickList__getNick(tmp->nicks, nick))
			strmcpy(list[idx++], tmp->channel, CHANNEL_LEN);

	return idx;
}

Winref	channel_window (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	if (ch)
		return Channel__window_get(ch);
	return 0;
}

void	channel_window_set (Winref win, Chname channel)
{
	Channel *ch = get_server_channel(channel, get_window_server(win));
	if (ch)
		return Channel__window_set(ch, win);
}

char *	channel_list_get (Servref server)
{
	Channel *	tmp = NULL;
	char *		retval = NULL;
	int		count = -1;

	while ((tmp = traverse_all_channels(&count, server)))
		if (tmp->status >= SYNCING && tmp->status <= PART_SENT)
			m_s3cat(&retval, space, tmp->channel);

	return retval ? retval : m_strdup(empty_string);
}

void	channel_window_move (Winref win)
{
	panic("Unimplemented");
}

void	channel_window_current_unset (Winref oldwin, Winref newwin)
{
	panic("Unimplemented");
}

void	channel_window_current_reset (Winref win)
{
	panic("Unimplemented");
}

Char *	channel_type (Chname channel, Servref server)
{
	panic("Unimplemented");
}

int	channel_syncing (Chname channel, Servref server)
{
	Channel *ch = get_server_channel(channel, server);
	if (ch && ch->status == SYNCING)
		return 1;

	return 0;
}

/* 
 * Im sure this doesnt belong here, but im not sure where it does belong.
 */
int 	auto_rejoin_callback (void *d)
{
	char *	data	= (char *) d;
	Chname 	channel	= next_arg(data, &data);
	Servref server	= my_atol(next_arg(data, &data));
	Channel *c;

	c = get_server_channel(channel, server);
	Channel__join(c);
	new_free((char **)&d);
	return 0;
}

/*
 * add_to_channel: adds the given nickname to the given channel.  If the
 * nickname is already on the channel, nothing happens.  If the channel is
 * not on the channel list, nothing happens (although perhaps the channel
 * should be addded to the list?  but this should never happen) 
 */
void 	channel_nick_add (Chname channel, Servref server, Nickname nickn, int suspicious, int chop, int voice, int helper)
{
	Nick 	*nick;
	Channel *chan;

	for (;;)
	{
		if (*nickn == '+')
			nickn++, voice = 1;
		else if (*nickn == '@')
			nickn++, chop = 1;
		else if (*nickn == '+')
			nickn++, helper = 1;
		/* Ignore OpenProject's" CanTopic" mode */
		else if (*nickn == '=')
			nickn++;
		else
			break;
	}

	if (!(chan = get_server_channel(channel, server)))
		return;

	if (is_me(server, nickn))
	{
		if (chop)
			chan->chop = 1;		/* Restore saved mode? */
		if (voice)
			chan->voice = 1;
		if (helper)
			chan->half_assed = 1;
	}

	nick = Nick__new(nickn, NULL, suspicious, chop, voice, helper);
	NickList__addNick(chan->nicks, nick);
}


/*
 * When a WHO reply comes back for a user, this gets called to add their
 * userhost to the nick structure
 */
void 	channel_userhost_add (Chname chan, Servref server, Nickname nickn, Char *uh)
{
	Nick 	*nick;

	/* XXX Need to "repair" nickname if possible here.  Now unimp */

	if ((nick = get_server_channel_nick(chan, server, nickn)))
		Nick__userhost_set(nick, uh);
}



/*
 * remove_from_channel: removes the given nickname from the given channel. If
 * the nickname is not on the channel or the channel doesn't exist, nothing
 * happens. 
 */
void 	channel_nick_remove (Chname channel, Servref server, Nickname nickname)
{
	Nick	*nick;
	Channel	*chan;

	if (channel == NULL)
	{
		int	count = -1;

		while ((chan = traverse_all_channels(&count, from_server)))
		{
		    if ((nick = NickList__getNick(chan->nicks, nickname)))
			if ((nick = NickList__removeNick(chan->nicks, nick)))
				Nick__delete(nick);
		}
	}
	else
	{
		if ((chan = get_server_channel(channel, server)))
		{
		    if ((nick = NickList__getNick(chan->nicks, nickname)))
			if ((nick = NickList__removeNick(chan->nicks, nick)))
				Nick__delete(nick);
		}
	}
}

/*
 * rename_nick: in response to a changed nickname, this looks up the given
 * nickname on all you channels and changes it the new_nick.  Now it also
 * restores the userhost (was lost before [oops!])
 */
void 	channel_nick_rename (Servref server, Nickname oldnick, Nickname newnick)
{
	ChannelList *	cl = get_server_channels(server);
	Channel *	chan;
	Nick *		nick;
	int		n = 0;

	while ((chan = ChannelList__getN(cl, n)))
	{
		if ((nick = NickList__getNick(chan->nicks, oldnick)))
		{
			NickList__removeNick(chan->nicks, nick);
			Nick__nickname_set(nick, newnick);
			NickList__addNick(chan->nicks, nick);
		}
		n++;
	}
}

int 	channel_nick_onchannel (Chname channel, Servref server, Nickname nick)
{
	return (get_server_channel_nick(channel, server, nick) ? 1 : 0);
}


int 	channel_nick_chanop (Chname channel, Servref server, Nickname nick)
{
	Nick *n;

	if ((n = get_server_channel_nick(channel, server, nick)))
		return n->chanop;
	else
		return 0;
}


int 	channel_nick_voice (Chname channel, Servref server, Nickname nick)
{
	Nick *n;

	if ((n = get_server_channel_nick(channel, server, nick)))
		return n->voice;
	else
		return 0;
}


int	channel_nick_number (Chname channel, Servref server)
{
	Channel *chan = get_server_channel(channel, server);

	if (channel)
		return chan->nicks->max;
	else
		return 0;
}


char	*channel_nick_list (Chname channel, Servref server)
{
	Channel *chan = get_server_channel(channel, server);
	char 	*str = NULL;
	int 	i;

	if (channel)
	{
		for (i = 0; i < chan->nicks->max; i++)
		{
		    Nick *nick = chan->nicks->list[i];
		    m_s3cat(&str, space, Nick__nickname_get(nick));
		}
	}
	return str;
}


char	*channel_nick_allchops (Chname channel, Servref server)
{
	Channel *chan = get_server_channel(channel, server);
	char 	*str = NULL;
	int 	i;

	if (channel)
	{
		for (i = 0; i < chan->nicks->max; i++)
		{
			Nick *nick = chan->nicks->list[i];
			if (Nick__chanop_get(nick))
			    m_s3cat(&str, space, Nick__nickname_get(nick));
		}
	}

	return str ? str : m_strdup(empty_string);
}


char	*channel_nick_allnochops (Chname channel, Servref server)
{
	Channel *chan = get_server_channel(channel, server);
	char 	*str = NULL;
	int 	i;

	if (channel)
	{
		for (i = 0; i < chan->nicks->max; i++)
		{
			Nick *nick = chan->nicks->list[i];
			if (!Nick__chanop_get(nick))
			    m_s3cat(&str, space, Nick__nickname_get(nick));
		}
	}

	return str ? str : m_strdup(empty_string);
}


Char *	fetch_userhost (Servref server, Nickname nickname)
{
	ChannelList *	cl = get_server_channels(server);
	Channel *	chan;
	Nick *		nick;
	int		n = 0;

	while ((chan = ChannelList__getN(cl, n)))
	{
		if ((nick = NickList__getNick(chan->nicks, nickname)))
			return nick->userhost;
		n++;
	}
	return NULL;
}


char *	scan_channel (Chname channel, Servref server)
{
	Channel *	wc = get_server_channel(channel, server);
	NickList *	nicks;
	char		buffer[NICKNAME_LEN + 5];
	char *		retval = NULL;
	int		i;

	if (!wc)
		return m_strdup(empty_string);

	nicks = wc->nicks;
	for (i = 0; i < nicks->max; i++)
	{
		if (nicks->list[i]->chanop)
			buffer[0] = '@';
		else
			buffer[0] = '.';

		if (nicks->list[i]->voice == 1)
			buffer[1] = '+';
		else if (nicks->list[i]->voice == -1)
			buffer[1] = '?';
		else
			buffer[1] = '.';

		strlcpy(buffer + 2, nicks->list[i]->nick, NICKNAME_LEN);
		m_s3cat(&retval, space, buffer);
	}

	return retval;
}



#if 0
/******************************* START PROCESSING HERE *******************/


/* This is a keybinding */
void 	switch_channels (char dumb, char *dumber)
{
	Channel *tmp = NULL;
	Channel *start = NULL;
	char	*nc = get_window_channel(0);

	if (!channel_list)
		return;		/* Dont bother */

	if (nc)
		if ((start = find_channel(nc, from_server)))
			tmp = start->next;

	if (!tmp)
		tmp = channel_list;

	/*
	 * This attempts to make an entire pass through the channel_list
	 * starting at the "previous" channel.
	 */
	while (tmp != start)
	{
		if (tmp->server != from_server || 
			is_current_channel(tmp->channel, 0))
		{
			tmp = tmp->next;
			if (!tmp && start)
				tmp = channel_list;
			continue;
		}

		set_window_channel(0, tmp->channel);
		update_all_windows();
		break;
	}
}

/*
 * XXX Rewrite
 */
void 	change_server_channels (int old_s, int new_s)
{
	Channel *tmp = NULL;

	TRAVERSE_ALL_CHANNELS(old_s, tmp)
	{
		tmp->server = new_s;
		clear_channel(tmp);
	}
}

/*
 * XXX Rewrite
 */
void 	destroy_server_channels (int server)
{
	Channel	*tmp = NULL;

	TRAVERSE_ALL_CHANNELS(server, tmp)
	{
		destroy_channel(tmp);
		new_free((char **)&tmp);
	}

/* This is very very bad.  Don't do this */
#if 0
	/* Try to figure out what channel to use... >;-) */
	if (!dead)
		switch_channels(0, 0);
#endif
}


/*
 * reconnect_all_channels: used after you get disconnected from a server, 
 * clear each channel nickname list and re-JOINs each channel in the 
 * channel_list ..  
 */
void 	reconnect_all_channels (void)
{
	Channel *tmp = NULL;
	char	*s;
	char 	*t;
	char	*channels = NULL;
	char	*keys = NULL;

	TRAVERSE_ALL_CHANNELS(from_server, tmp)
	{
		if ((s = Channel__mode_string(tmp)))
			malloc_strcpy(&tmp->cached_mode, s);

		m_s3cat(&channels, ",", tmp->channel);
		/* Ugh.  Hybrid made me do it. */
		m_s3cat(&keys, ",", tmp->key ? tmp->key : empty_string);

		clear_channel(tmp);
	}

	/*
	 * Interestingly enough, black magic on the server's part makes
	 * this work.  I sure hope they dont "break" this in the future...
	 */
	if (channels)
		send_to_server("JOIN %s %s", channels, 
						keys ? keys : empty_string);

	new_free(&channels);
	new_free(&keys);
	destroy_server_channels(from_server);
}


const char 	*what_channel (const char *nick)
{
	Channel *tmp = NULL;
	int	count = -1;

	if (current_window->current_channel &&
	    channel_nick_onchannel(current_window->current_channel, 
				   current_window->server, nick))
		return current_window->current_channel;

	while ((tmp = traverse_all_channels(&count, from_server)))
	{
		if (NickList__getNick(tmp->nicks, nick))
			return tmp->channel;
	}

	return NULL;
}

const char 	*walk_channels (int init, const char *nick)
{
	static	Channel *tmp = (Channel *) 0;
	int	count = -1;

	if (init)
		tmp = NULL;

	while ((tmp = traverse_all_channels(&count, from_server)))
	{
		if (NickList__getNick(tmp->nicks, nick))
			return tmp->channel;
	}

	return NULL;
}

int 	get_channel_oper (const char *channel, int server)
{
	Channel *chan;

	if ((chan = find_channel(channel, server)))
		return chan->chop;
	else
		return 0;
}

int 	get_channel_voice (const char *channel, int server)
{
	Channel *chan;

	if ((chan = find_channel(channel, server)))
		return chan->voice;
	else
		return 0;
}

Winref	get_channel_window (const char *channel, int server)
{
	Channel *tmp = find_channel(channel, server);

	if (tmp)
		return tmp->window;

	return NULL;
}

void 	set_channel_window (Winref window, const char *channel)
{
	Channel *tmp = find_channel(channel, window->server);

	if (tmp)
		tmp->window = window;
}

/*
 * This finds the channel that is the current channel of the specified
 * window, and sets the channel to some new window that wants to steal it.
 */
void	unset_window_current_channel (Winref old_w, Winref new_w)
{
	Channel *tmp = NULL;
	int	count = -1;

	while ((tmp = traverse_all_channels(&count, old_w->server)))
	{
		if (tmp->window == old_w && !my_stricmp(tmp->channel, old_w->current_channel))
		{
			new_free(&old_w->current_channel);
			tmp->window = new_w;
			return;
		}
	}
}

/*
 * This attempts to find another current channel for a window when the 
 * current channel for a window is 'stolen' by another window.
 */
void	reset_window_current_channel (Winref w)
{
	Channel *tmp = NULL;
	int	count = -1;

	while ((tmp = traverse_all_channels(&count, w->server)))
	{
		if (tmp->window == w)
		{
			malloc_strcpy(&w->current_channel, tmp->channel);
			return;
		}
	}
}

/*
 * For any given window, re-assign all of the channels that are connected
 * to that window.
 */
void	move_window_channels (Winref window)
{
	Channel *tmp = NULL;
	Winref *w = NULL;
	int	count = -1;

	while ((tmp = traverse_all_channels(&count, window->server)))
	{
		if (tmp->window != window)
			continue;

		tmp->window = NULL;
		while (traverse_all_windows_logical(&w))
		{
			if (get_window_server(w) == tmp->server && 
				w != window)
			{
				tmp->window = w;
				break;
			}
		}
	}
}

char *	create_channel_list (int server)
{
	Channel	*tmp = NULL;
	char	*retval = NULL;
	int	count = -1;

	while ((tmp = traverse_all_channels(&count, server)))
		m_s3cat(&retval, space, tmp->channel);

	return retval ? retval : m_strdup(empty_string);
}

void 	channel_server_delete (int i)
{
	Channel	*tmp = NULL;
	int	count = -1;

	while ((tmp = traverse_all_channels(&count, i)))
		if (tmp->server >= i)
			tmp->server--;
}

#endif

/*
 * Converts a raw channel name into a more user-friendly form. Right now
 * all we support are ircnet's serial-numbered channels.
 */
char *	make_friendly (Char *channel)
{
	char	stuff[IRCD_BUFFER_SIZE + 1];

	/* IRCNet serial-numbered channels */
	if (*channel == '!' && strlen(channel) >= 6)
	{
		snprintf(stuff, IRCD_BUFFER_SIZE, "[%.6s] %s",
				channel, channel + 6);
	}

	/* Everybody else */
	else
		strcpy(stuff, channel);

	return m_strdup(stuff);
}

void 	add_user_who (char *from, char **ArgList)
{
	char *userhost;

	if (x_debug & DEBUG_CHANNELS)
		yell("Add_user_who!");

	userhost = alloca(strlen(ArgList[1]) + strlen(ArgList[2]) + 2);
	strcpy(userhost, ArgList[1]);
	strcat(userhost, "@");
	strcat(userhost, ArgList[2]);
	channel_userhost_add(ArgList[0], from_server, ArgList[4], userhost);
}

void 	add_user_end (char *from, char **ArgList)
{
	if (x_debug & DEBUG_CHANNELS)
		yell("Add_user_end!");

	channel_status_change(ArgList[0], from_server, ACTIVE);
}

BUILT_IN_COMMAND(partcmd)
{
	char	*cname;
	char	*reason;
	Channel *chan;

	if (!(cname = next_arg(args, &args)))
		cname = empty_string;
	if (!args || !*args)
		args = NULL;
	reason = args;

	if (!cname || !*cname || !strcmp(cname, "*"))
	{
		cname = get_window_channel(0);
		if (!cname || !*cname)
			cname = "*";		/* What-EVER */
	}

	if ((chan = get_server_channel(cname, from_server)))
		Channel__part(chan, reason);
	else
		say("You are not on that channel!");
}

